/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    segEdgeLabeler.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**
** Keyboard:
**  TAB cycles through labels
**  'v' cycles through views
**  SHIFT toggles convex/concave, ccclusion left / occlusion right
**
*****************************************************************************/

#include "segEdgeLabeler.h"

#if defined(_WIN32)||defined(WIN32)||defined(__WIN32__)
#define _CRT_SECURE_NO_DEPRECATE
#undef max
#undef min
#endif

#include <cstdlib>
#include <cstdio>
#include <cmath>
#include <limits>
#include <iostream>
#include <fstream>
#include <deque>


#include "wx/wx.h"
#include "wx/utils.h"
#include "wx/wxprec.h"
#include "wx/cmdline.h"
#include "wx/aboutdlg.h"


using namespace std;

#define NOT_IMPLEMENTED_YET wxMessageBox(_T("Functionality not implementet yet."),\
	_T("Error"), wxOK | wxICON_EXCLAMATION, this);

// Global Variables and Tables -------------------------------------------------

MainWindow *gMainWindow = NULL;

static const wxCmdLineEntryDesc COMMAND_LINE_DESCRIPTION[] =
{
    { wxCMD_LINE_OPTION, "i", "image", "image" },
    { wxCMD_LINE_OPTION, "s", "segmentation", "segmentation directory" },

    { wxCMD_LINE_NONE }
};

#define NUM_LABELS 6

const char *LABEL_NAMES[NUM_LABELS] = {
    "<unknown>",
    "phantom",
    "concave fold",
    "convex fold",
    "occlusion (ccw)",
    "occlusion (cw)"
};

const unsigned char LABEL_COLOURS[NUM_LABELS][3] = {
    {0x00, 0x00, 0x00},
    {0x00, 0x00, 0xff},
    {0x7f, 0x00, 0x00},
    {0x00, 0x7f, 0x00},
    {0xaf, 0x1f, 0x7f},
    {0x7f, 0x1f, 0xaf}
};

string gSuperpixelDir;

// Event Tables ----------------------------------------------------------------

BEGIN_EVENT_TABLE(MainCanvas, wxWindow)
    EVT_ERASE_BACKGROUND(MainCanvas::on_erase_background)
    EVT_SIZE(MainCanvas::on_size)
    EVT_PAINT(MainCanvas::on_paint)
    EVT_CHAR(MainCanvas::on_key)
    EVT_MOUSE_EVENTS(MainCanvas::on_mouse)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(MainWindow, wxFrame)
    EVT_CLOSE(MainWindow::on_close)

    EVT_MENU(FILE_NEW, MainWindow::on_file_menu)
    EVT_MENU(FILE_OPEN, MainWindow::on_file_menu)
    EVT_MENU(FILE_OPEN_IMAGE, MainWindow::on_file_menu)
    EVT_MENU(FILE_SAVE, MainWindow::on_file_menu)
    EVT_MENU(FILE_SAVEAS, MainWindow::on_file_menu)
    EVT_MENU(FILE_EXIT, MainWindow::on_file_menu)

    EVT_MENU(EDIT_SET_UNKNOWN, MainWindow::on_edit_menu)
    EVT_MENU(EDIT_CHANGE_LABEL, MainWindow::on_edit_menu)
    EVT_MENU(EDIT_UNDO, MainWindow::on_edit_menu)

    EVT_MENU(OPTIONS_VIEW_IMAGE, MainWindow::on_options_menu)
    EVT_MENU(OPTIONS_VIEW_EDGES, MainWindow::on_options_menu)
    EVT_MENU(OPTIONS_VIEW_EDGE_OVERLAY, MainWindow::on_options_menu)
    EVT_MENU(OPTIONS_VIEW_REALEDGE_OVERLAY, MainWindow::on_options_menu)

    EVT_TOOL_RANGE(TOOLBAR_BASE, TOOLBAR_BASE + 1000, MainWindow::on_toolbar)

    EVT_MENU(HELP_ABOUT, MainWindow::on_help_menu)
END_EVENT_TABLE()

// MainCanvas Implementation ---------------------------------------------------

MainCanvas::MainCanvas(wxWindow *parent, wxWindowID id,const wxPoint& pos,
    const wxSize& size, long style, const wxString& name) :
    wxWindow(parent, id, pos, size, style, name),
    _image(NULL), _edgeSnapMap(NULL), _imageBuffer(NULL),
    _activeLabel(0), _viewMode(VM_EDGE_OVERLAY)
{
    SetMinSize(wxSize(320, 240));
    updateToolBar();
}

MainCanvas::~MainCanvas()
{
    if (_image != NULL) {
        delete _image;
        cvReleaseMat(&_edgeSnapMap);
        delete[] _imageBuffer;
    }

    // TO DO: warn if not saved
}

void MainCanvas::on_erase_background(wxEraseEvent &event)
{
    // do nothing (and avoid flicker)
}

void MainCanvas::on_paint(wxPaintEvent &WXUNUSED(event))
{
    int width, height;
    GetClientSize(&width, &height);

    wxPaintDC dc(this);
    if (_image == NULL) {
        dc.Clear();
        dc.SetTextForeground(wxColor(0, 0, 255));
        wxSize s = dc.GetTextExtent("no image");
        dc.DrawText("no image", (int)(width - s.x)/2, (int)(height - s.y)/2);
	    return;
    }

    // copy frame to screen
    wxImage screenImage(_image->width(), _image->height());
    screenImage.SetData(_imageBuffer, true);
    dc.DrawBitmap(screenImage.Scale(width, height), 0, 0);
}

void MainCanvas::on_size(wxSizeEvent &event)
{
    int width, height;

    GetClientSize(&width, &height);

    this->Refresh(false);
    this->Update();
}

void MainCanvas::on_key(wxKeyEvent &event)
{
    switch (event.m_keyCode) {
    case WXK_DELETE:
	// TO DO: clear current region
	break;
    case WXK_TAB:
    {
	// TO DO: shift goes backwards
        setActiveLabel((_activeLabel + 1) % NUM_LABELS);
	break;
    }
    case WXK_SHIFT:
    {
        if (_activeLabel == 2) setActiveLabel(3);
        else if (_activeLabel == 3) setActiveLabel(2);
        else if (_activeLabel == 4) setActiveLabel(5);
        else if (_activeLabel == 5) setActiveLabel(4);
    }
    case 'v':
    case 'V':
	if (_viewMode != VM_REALEDGE_OVERLAY) {
	    setViewMode((TViewMode)(_viewMode + 1));
	} else {
	    setViewMode(VM_IMAGE);
	}
	break;
    default:
    	event.Skip();
    }

    // refresh view
    updateStatusBar();
    this->Refresh(false);
    this->Update();
}

void MainCanvas::on_mouse(wxMouseEvent &event)
{
    if (_image == NULL)
	return;

    int width, height;
    GetClientSize(&width, &height);
    double scaleX = (double)_image->width() / (double)width;
    double scaleY = (double)_image->height() / (double)height;

    if (event.LeftDown() || event.Dragging()) {
        int x = (int)(scaleX * event.m_x);
        int y = (int)(scaleY * event.m_y);
        int indx = -1;
        if ((x >= 0) && (y >= 0) && 
            (x < _image->width()) && (y < _image->height()))
            indx = findNearestEdge(x, y);
        if (indx >= 0) {
	    undoableAction();
            bool flip = false;
            _labels[indx] = _activeLabel;
            
            if ((_activeLabel == 4) || (_activeLabel == 5)) {
                int indxA = _adjList[indx].first;
                int indxB = _adjList[indx].second;
                double dx = _image->getCentroid(indxA).x - _image->getCentroid(indxB).x;
                double dy = _image->getCentroid(indxB).y - _image->getCentroid(indxA).y;
                
                double theta = atan2(dy, dx);
                if ((theta > -M_PI / 2) && (theta < M_PI / 2))
                    flip = true;
            }

            if (event.ShiftDown())
                flip = !flip;

            if (flip) {
                if ((_activeLabel == 2) || (_activeLabel == 4)) {
                    _labels[indx] = _activeLabel + 1;
                } else if ((_activeLabel == 3) || (_activeLabel == 5)) {
                    _labels[indx] = _activeLabel - 1;
                }
            }
            updateImageBuffer();
	}
    }

    event.Skip();
    updateStatusBar();
    this->Refresh(false);
    this->Update();
}

void MainCanvas::setViewMode(TViewMode m)
{
    _viewMode = m;
    updateImageBuffer();
    updateStatusBar();
}

void MainCanvas::setActiveLabel(int lbl)
{
    _activeLabel = lbl;
    updateStatusBar();
}

void MainCanvas::setAllUnknown()
{
    undoableAction();
    for (int i = 0; i < (int)_labels.size(); i++) {
        if (_labels[i] == 0)
            _labels[i] = 1;
    }
    updateImageBuffer();
    updateStatusBar();
}

bool MainCanvas::openImage(const char *filename)
{
    assert(filename != NULL);

    if (_image != NULL) {
        delete _image;
        cvReleaseMat(&_edgeSnapMap);
        delete[] _imageBuffer;
    }

    string imgFilename = string(filename);
    string segFilename = strBaseName(imgFilename) + string(".seg.int");
    if (gSuperpixelDir.empty())
        segFilename = strDirectory(imgFilename) + string("/../superpixels/") +
            segFilename;
    else segFilename = gSuperpixelDir + string("/") + segFilename;
    
    _image = new svlSegImageBase(imgFilename.c_str(), segFilename.c_str());
    assert(_image != NULL);
    _adjList = _image->getAdjacencyList();
    _edgels = _image->getEdgels();
    _edgeSnapMap = cvCreateMat(_image->height(), _image->width(), CV_32SC1);
    cvSet(_edgeSnapMap, cvScalarAll(-2));
    for (int i = 0; i < (int)_edgels.size(); i++) {
        for (int j = 0; j < (int)_edgels[i].size(); j++) {
            CV_MAT_ELEM(*_edgeSnapMap, int, _edgels[i][j].y, _edgels[i][j].x) = i;
        }
    }

    _imageBuffer = new unsigned char[3 * _image->width() * _image->height()];
    newLabels();

    updateImageBuffer();
    return true;
}

void MainCanvas::newLabels()
{
    _labels.clear();
    if (_image != NULL) {
        _labels.resize(_edgels.size(), 0);
    }
    _undoLabels = _labels;

    updateImageBuffer();
}

bool MainCanvas::openLabels(const char *filename)
{
    assert(filename != NULL);
    if (_image == NULL)
        return false;

    newLabels();

    ifstream ifs(filename);
    assert(!ifs.fail());
    int id;
    for (unsigned i = 0; i < _labels.size(); i++) {
        ifs >> id;
        _labels[i] = id + 1;
        assert((_labels[i] >= 0) && (_labels[i] < NUM_LABELS));
    }
    ifs.close();

    updateImageBuffer();
    return true;
}

bool MainCanvas::saveLabels(const char *filename)
{
    assert(filename != NULL);
    if (_image == NULL)
        return false;
    
    ofstream ofs(filename);
    assert(!ofs.fail());
    for (unsigned i = 0; i < _labels.size(); i++) {
        ofs << (_labels[i] - 1) << "\n";
    }
    ofs.close();

    return true;
}

void MainCanvas::undo()
{
    vector<int> _oldLabels(_labels);
    _labels = _undoLabels;
    _undoLabels = _oldLabels;
    
    updateImageBuffer();
}

int MainCanvas::findNearestEdge(int x, int y)
{    
    if (CV_MAT_ELEM(*_edgeSnapMap, int, y, x) != -2)
        return CV_MAT_ELEM(*_edgeSnapMap, int, y, x);

    int bestIndx = -1;
    int bestDistance = 4;

    for (int i = 0; i < (int)_edgels.size(); i++) {
        for (int j = 0; j < (int)_edgels[i].size(); j++) {
            int d = abs(_edgels[i][j].x - x) +
                abs(_edgels[i][j].y - y);
            if (d < bestDistance) { 
                bestIndx = i;
                bestDistance = d;
            }
        }
    }

    CV_MAT_ELEM(*_edgeSnapMap, int, y, x) = bestIndx;
    return bestIndx;
}

void MainCanvas::undoableAction()
{
    _undoLabels = _labels;
}

void MainCanvas::updateImageBuffer()
{
    if (_imageBuffer == NULL)
	return;

    // copy image to buffer
    const IplImage *img = _image->image();
    for (int y = 0; y < _image->height(); y++) {
        for (int x = 0; x < _image->width(); x++) {
            _imageBuffer[3 * (y * img->width + x) + 0] =
                img->imageData[y * img->widthStep + 3 * x + 2];
            _imageBuffer[3 * (y * img->width + x) + 1] =
                img->imageData[y * img->widthStep + 3 * x + 1];
            _imageBuffer[3 * (y * img->width + x) + 2] =
                img->imageData[y * img->widthStep + 3 * x + 0];
        }
    }

    if (_viewMode == VM_IMAGE)
        return;

    if (_viewMode == VM_EDGES) {
        memset(_imageBuffer, 0, _image->width() * _image->height() * 3);
    }

    for (unsigned i = 0; i < _edgels.size(); i++) {

        if ((_viewMode == VM_REALEDGE_OVERLAY) && (_labels[i] < 2))
            continue;

        // colour foreground
        if ((_labels[i] == 4) || (_labels[i] == 5)) {
            for (unsigned j = 0; j < _edgels[i].size(); j++) {
                int x = _edgels[i][j].x;
                int y = _edgels[i][j].y;
                int segId = _adjList[i].first;
                if (_labels[i] == 5) segId = _adjList[i].second;
                for (int yy = y - 2; yy <= y + 2; yy++) {
                    if ((yy < 0) || (yy >= _image->height())) continue;
                    for (int xx = x - 2; xx <= x + 2; xx++) {
                        if ((xx < 0) || (xx >= _image->width())) continue;
                        if (_image->getSegment(xx, yy) == segId) {
                            colourBufferPixel(xx, yy, 0xff, 0xff, 0xff);
                        }
                    }
                }
            }
        }

        // color edgel
        for (unsigned j = 0; j < _edgels[i].size(); j++) {
            int x = _edgels[i][j].x;
            int y = _edgels[i][j].y;
            int thickness = 1;
            if ((_labels[i] == 0) || (_labels[i] == 1)) thickness = 0;
            for (int yy = y - thickness; yy <= y + thickness; yy++) {
                if ((yy < 0) || (yy >= _image->height())) continue;
                for (int xx = x - thickness; xx <= x + thickness; xx++) {
                colourBufferPixel(xx, yy, LABEL_COLOURS[_labels[i]][0],
                    LABEL_COLOURS[_labels[i]][1], LABEL_COLOURS[_labels[i]][2]);
                }
            }
        }
    }
}

void MainCanvas::colourBufferPixel(int x, int y,
    unsigned char red, unsigned char green, unsigned char blue)
{
    _imageBuffer[3 * (y * _image->width() + x) + 0] = red;
    _imageBuffer[3 * (y * _image->width() + x) + 1] = green;
    _imageBuffer[3 * (y * _image->width() + x) + 2] = blue;
}

void MainCanvas::updateStatusBar()
{
    ((MainWindow *)GetParent())->SetStatusText(wxString::Format(
        "Current label: %d (%s)", _activeLabel - 1, LABEL_NAMES[_activeLabel]));
}

void MainCanvas::updateToolBar()
{
    wxToolBar *toolbar = ((wxFrame *)this->GetParent())->GetToolBar();
    toolbar->ClearTools();
    toolbar->SetToolBitmapSize(wxSize(24, 24));

    wxMemoryDC memDC;
    wxBitmap *bmp; // reference counted so will be deleted automatically

    for (int i = 0; i < NUM_LABELS; i++) {
        bmp = new wxBitmap(24, 24);
        memDC.SelectObject(*bmp);
        memDC.SetBrush(wxBrush(wxColour(LABEL_COLOURS[i][0],
            LABEL_COLOURS[i][1], LABEL_COLOURS[i][2])));
        memDC.DrawRectangle(0, 0, 24, 24);
        toolbar->AddTool(TOOLBAR_BASE + i, _T(LABEL_NAMES[i]), *bmp,
            _T(LABEL_NAMES[i]));
    }
    
#if defined(_WIN32)||defined(WIN32)||defined(__WIN32__)
    // win32 workaround
    bmp = new wxBitmap(24, 24);
    memDC.SelectObject(*bmp);
#endif

    toolbar->Realize();
}

// MainWindow Implementation ---------------------------------------------------

MainWindow::MainWindow(wxWindow* parent, wxWindowID id, const wxString& title,
    const wxPoint& pos, const wxSize& size, long style) : 
    wxFrame(parent, id, title, pos, size, style)
{
    wxMenu *file_menu = new wxMenu;
    wxMenu *edit_menu = new wxMenu;
    wxMenu *options_menu = new wxMenu;
    wxMenu *help_menu = new wxMenu;

    file_menu->Append(FILE_NEW, _T("&New\tCtrl-N"), _T("Clear labels"));
    file_menu->Append(FILE_OPEN, _T("&Open...\tCtrl-O"), _T("Open labels"));
    file_menu->Append(FILE_OPEN_IMAGE, _T("Open &Image...\tCtrl-I"), _T("Open image and oversegmentation"));
    file_menu->AppendSeparator();
    file_menu->Append(FILE_SAVE, _T("&Save\tCtrl-S"), _T("Save labels"));
    file_menu->Append(FILE_SAVEAS, _T("Save &As..."), _T("Save labels"));
    file_menu->AppendSeparator();
    file_menu->Append(FILE_EXIT, _T("E&xit\tAlt-X"), _T("Exit this program"));
    edit_menu->Append(EDIT_SET_UNKNOWN, _T("&Set unlabeled as phantom"), _T("Sets the value of all unlabled edges to phantom"));
    edit_menu->AppendSeparator();
    edit_menu->Append(EDIT_CHANGE_LABEL, _T("&Change label..."), _T("Change current label"));
    edit_menu->AppendSeparator();
    edit_menu->Append(EDIT_UNDO, _T("&Undo\tCtrl-Z"), _T("Undo last change"));
    options_menu->AppendRadioItem(OPTIONS_VIEW_IMAGE, _T("View &image"), _T("Set viewing mode"));
    options_menu->AppendRadioItem(OPTIONS_VIEW_EDGES, _T("View &edges"), _T("Set viewing mode"));
    options_menu->AppendRadioItem(OPTIONS_VIEW_EDGE_OVERLAY, _T("View edge &overlay"), _T("Set viewing mode"));
    options_menu->AppendRadioItem(OPTIONS_VIEW_REALEDGE_OVERLAY, _T("View &folds/occlusions"), _T("Set viewing mode"));
    help_menu->Append(HELP_ABOUT, _T("&About...\tF1"), _T("Show about dialog"));

    wxMenuBar *menu_bar = new wxMenuBar();
    menu_bar->Append(file_menu, _T("&File"));
    menu_bar->Append(edit_menu, _T("&Edit"));
    menu_bar->Append(options_menu, _T("&Options"));
    menu_bar->Append(help_menu, _T("&Help"));
    SetMenuBar(menu_bar);

    options_menu->Check(OPTIONS_VIEW_EDGE_OVERLAY, true);

    this->CreateStatusBar();
    this->CreateToolBar(wxNO_BORDER | wxHORIZONTAL | wxTB_FLAT);

    // this is required for keyboard focus under Linux
    _canvas = new MainCanvas(this);
    _canvas->setViewMode(VM_EDGE_OVERLAY);
}

MainWindow::~MainWindow()
{
    // do nothing
}

void MainWindow::on_file_menu(wxCommandEvent& event)
{
    if (event.GetId() == FILE_NEW) {
        wxMessageDialog dlg(this, _T("Clear all labels?"),
            _T("New"), wxYES_NO | wxICON_QUESTION);
        if (dlg.ShowModal() == wxID_YES) {
            _canvas->newLabels();
            _labelFilename = string("");
	    SetTitle("SegEdgeLabeler");
        }
    } else if (event.GetId() == FILE_OPEN) {
        wxFileDialog dlg(this, _T("Choose label file to open"), _T(""), 
	    _T(""), _T("Text files (*.txt)|*.txt|All files (*.*)|*.*"), wxOPEN | wxFD_CHANGE_DIR);
        if (dlg.ShowModal() == wxID_OK) {
            _labelFilename = dlg.GetPath();
	    _canvas->openLabels(_labelFilename.c_str());
	    SetTitle((string("SegEdgeLabeler [") +
                         dlg.GetFilename() + string("]")).c_str());
        }
    } else if (event.GetId() == FILE_OPEN_IMAGE) {
        wxFileDialog dlg(this, _T("Choose image to open"), _T(""), 
	    _T(""), _T("Image files (*.jpg)|*.jpg|All files (*.*)|*.*"), wxOPEN | wxFD_CHANGE_DIR);
        if (dlg.ShowModal() == wxID_OK) {
            _canvas->openImage(dlg.GetPath().c_str());
	    _labelFilename.clear();
	    SetTitle((string("SegEdgeLabeler [") +
                         dlg.GetFilename() + string("]")).c_str());
        }
    } else if ((event.GetId() == FILE_SAVE) && (_labelFilename.size())) {
	_canvas->saveLabels(_labelFilename.c_str());
    } else if ((event.GetId() == FILE_SAVE) || (event.GetId() == FILE_SAVEAS)) {
        wxFileDialog dlg(this, _T("Choose label file to save"), _T(""), 
	    _T(""), _T("Text files (*.txt)|*.txt"), wxSAVE | wxFD_CHANGE_DIR);
        if (dlg.ShowModal() == wxID_OK) {
	    _labelFilename = dlg.GetPath().c_str();
	    _canvas->saveLabels(_labelFilename.c_str());
	    SetTitle((string("SegEdgeLabeler [") +
                    dlg.GetFilename() + string("]")).c_str());
        }
    } else if (event.GetId() == FILE_EXIT) {
        Close(true);
    }

    Refresh(false);
    Update();
}

void MainWindow::on_edit_menu(wxCommandEvent& event)
{
    if (event.GetId() == EDIT_SET_UNKNOWN) {
        _canvas->setAllUnknown();
    } if (event.GetId() == EDIT_CHANGE_LABEL) {
        wxTextEntryDialog dlg(this, "Enter label id:");
        if (dlg.ShowModal() == wxID_OK) {
	    _canvas->setActiveLabel(atoi(dlg.GetValue().c_str()));
        }
    } else if (event.GetId() == EDIT_UNDO) {
	_canvas->undo();
    }

    Refresh(false);
    Update();
}

void MainWindow::on_options_menu(wxCommandEvent& event)
{
    if (event.GetId() == OPTIONS_VIEW_IMAGE) {
	_canvas->setViewMode(VM_IMAGE);
    } else if (event.GetId() == OPTIONS_VIEW_EDGES) {
	_canvas->setViewMode(VM_EDGES);
    } else if (event.GetId() == OPTIONS_VIEW_EDGE_OVERLAY) {
	_canvas->setViewMode(VM_EDGE_OVERLAY);
    } else if (event.GetId() == OPTIONS_VIEW_REALEDGE_OVERLAY) {
	_canvas->setViewMode(VM_REALEDGE_OVERLAY);
    }

    Refresh(false);
    Update();
}

void MainWindow::on_help_menu(wxCommandEvent& event)
{
    if (event.GetId() == HELP_ABOUT) {
        wxAboutDialogInfo info;
        info.SetName(_("SegEdgeLabeler"));
        info.SetVersion(_("0.1"));
        info.SetDescription(_("This program allows you to label edges in an oversegmented image."));
        info.SetCopyright(_T("(C) 2008 Staphen Gould <sgould@stanford.edu>"));

        wxAboutBox(info);
    }
}

void MainWindow::on_toolbar(wxCommandEvent& event)
{
    int toolId = event.GetId() - (int)TOOLBAR_BASE;
    _canvas->setActiveLabel(toolId);
}

void MainWindow::on_close(wxCloseEvent& event)
{
    // not implemented yet
    event.Skip();
}

// SegEdgeLabelerApp Implementation --------------------------------------------

bool SegEdgeLabelerApp::OnInit()
{
    // setup main window
    gMainWindow = new MainWindow(NULL, wxID_ANY, wxT("SegEdgeLabeler"),
        wxDefaultPosition, wxSize(640, 480));
    SetTopWindow(gMainWindow);
    gMainWindow->Show();
    gMainWindow->SetFocus();    

    // call base class for command-line options
    wxApp::OnInit();

    return true;
}

void SegEdgeLabelerApp::OnInitCmdLine(wxCmdLineParser& parser)
{
    parser.SetDesc(COMMAND_LINE_DESCRIPTION);
    if (parser.Parse(true)) {
	exit(1);
    }

    wxString str;
    if (parser.Found("i", &str)) {
	gMainWindow->_canvas->openImage(str.c_str());
    }
    if (parser.Found("s", &str)) {
        gSuperpixelDir = string(str.c_str());
    }
}

int SegEdgeLabelerApp::OnExit()
{
    return 0;
}

IMPLEMENT_APP(SegEdgeLabelerApp)

