/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    combineDetections.cpp
** AUTHOR(S):   Sid Batra <sidbatra@cs.stanford.edu>
**              Olga Russakovsky <olga@cs.stanford.edu>
**
** DESCRIPTION:
**  Combined results from different object classifiers.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>
#include <string.h>

#include "cv.h"
#include "cxcore.h"
#include "highgui.h"

#include "svlBase.h"
#include "svlML.h"
#include "svlVision.h"

using namespace std;

#define WINDOW_NAME "classifyImages"

// main -----------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << endl;
    cerr << "USAGE: ./combineDetections [OPTIONS] <groundTruth filename> <outputFilename> (<detection filenames>)" << endl;
    cerr << "OPTIONS:" << endl
	 << "  -p <filename>         :: projection of detections to file" << endl
         << "  -s <filename>         :: source folder where the original depth maps are kept" << endl
	 << endl;
}


#define NUM_REQUIRED_PARAMETERS 3

int main(int argc, char *argv[])
{		
       
    // read commandline parameters
    const char *projectionFilename = NULL;
    const char *sourceFolder = NULL;
    const char *outputFilename = NULL;
    const char *groundTruthFilename = NULL;
   
    char **args = argv + 1;
    while (--argc) {
        if (!strcmp(*args, "-p")) {
            projectionFilename = *(++args); argc--;
        } else if (!strcmp(*args, "-s")) {
            sourceFolder = *(++args); argc--;
        } else if (**args == '-') {
            SVL_LOG(SVL_LOG_FATAL, "unrecognized option " << *args);
        } else {
            break;
        }
        args++;
    }

    if (argc < NUM_REQUIRED_PARAMETERS) 
    {
    	usage();
    	return -1;
    }

    groundTruthFilename = args[0];
    outputFilename = args[1];
    
    vector<svlObject2dSequence> detections;
    svlObject2dSequence groundTruth;

    cout<<"\nReading ground truth file\n";

    readObject2dSequence(groundTruthFilename,groundTruth);

    cout<<"\n Reading various detection sequences\n";

    for (unsigned i = 2; i < (unsigned)(argc); i++) 
    {        
        cout<<"\n Reading sequence number "<<i-2;
        
        svlObject2dSequence objectList;
        readObject2dSequence(args[i],objectList);
        detections.push_back(objectList);
    }

    // TODO: make this much more general    

    //Used to replace names in combined sequence to none so that nonMaximalSupression may work
    int numMaps = 7;
    string stringMap[] = {"mug"  , "cup" , "stapler", "clock", "monitor", "ski_boot", "door_handle"};
    int indexMap[] = {0 , 1, 2, 3, 4, 5, 6};

    svlObject2dSequence combinedSequence;
    svlObject2dSequence finalSequence;

        
    //Iterate through each image in the ground truth set
    for (svlObject2dSequence::const_iterator i = groundTruth.begin(); i != groundTruth.end(); ++i)
    {	

        cout<<"\n Processing image "<<i->first;

        //Holds the combined detections from all files in this frame
        svlObject2dFrame combined;

        //Combine detections from each of the different files
        for( unsigned j=0 ; j<detections.size() ; j++)
        {
            svlObject2dSequence::const_iterator k;
            bool isFound = false;

            for( k=detections[j].begin() ; k != detections[j].end() ; ++k)
            {
                if( k->first == i->first )
                {
                    isFound = true;
                    break;
                }
            }

            //Attempt to add frame data to combination only if detections are found in that frame
            //by the current classifier
            if( isFound )
            {
                for( unsigned l=0 ; l<k->second.size() ; l++)
                {
                    svlObject2d temp(k->second[l]);

		    bool nameFound = false;
		    for (int n = 0; n < numMaps; n++) {
		      if (temp.name == stringMap[n]) {
			temp.index = indexMap[n];
			nameFound = true;
			break;
		      }
		    }

		    SVL_ASSERT_MSG(nameFound, "groundtruth file contains unknown objects");

                    temp.name = string("none");
                    combined.push_back(temp);

                }
            }
            
        }

        
        combinedSequence[i->first] = combined;
        
    }

    //Suppress overlapping regions from different classifers to remove some false positives
    int removed = nonMaximalSuppression(combinedSequence, 0.25, 0.25, 0.75);
    SVL_LOG(SVL_LOG_MESSAGE, "removed " << removed << " overlapping objects");

    //Remap all the indices back to their object names
    for (svlObject2dSequence::const_iterator i = combinedSequence.begin(); i != combinedSequence.end(); ++i)
    {	
        svlObject2dFrame frame;
        for( unsigned k=0 ; k<i->second.size() ; k++)
        {
            svlObject2d temp(i->second[k]);
            
	    SVL_ASSERT_MSG(temp.index >= 0 && temp.index < numMaps, "for some reason index out of bounds");
	   
	    temp.name = stringMap[temp.index];

            //if( temp.index == indexMap[0] )
	    //   temp.name = stringMap[0];
            //else if( temp.index == indexMap[1] )
	    //   temp.name = stringMap[1];
            //else if( temp.index == indexMap[2] )
	    //   temp.name = stringMap[2];    

            frame.push_back(temp);
        }

        finalSequence[i->first] = frame;

    }

    writeObject2dSequence(outputFilename,finalSequence);

    //If projections of detections are to be stored
    if( projectionFilename != NULL && sourceFolder != NULL )
    {
        cout<<"\n Processing and writing projections of detections to file\n";

        //Open stream for writing projections
        ofstream out(projectionFilename,ios::out);

        out<<finalSequence.size()<<"\n";
        
        for (svlObject2dSequence::const_iterator i = finalSequence.begin(); i != finalSequence.end(); ++i)
        {
            cout<<"\n Processing image "<<i->first;

            out<<i->second.size()<<"\n";

            string filename = string(sourceFolder) + string("/") + i->first + ".reconstructed.z.txt";
   
            IplImage *depthMap = readMatrixAsIplImage(filename.c_str(),640,480);
            
            for( unsigned j=0 ; j<i->second.size() ; j++)
            {
                double z = CV_IMAGE_ELEM(depthMap,float,(int)(i->second[j].y +i->second[j].h/2),(int)(i->second[j].x+i->second[j].w/2));
                out<<i->second[j].name<<" "<<i->second[j].x+i->second[j].w/2<<" "
                    <<i->second[j].y +i->second[j].h/2<<" "<<z<<" "<<i->second[j].pr;

                if( j != i->second.size() -1 )
                    out<<"\n";
            }

            out<<"\n";

            cvReleaseImage(&depthMap);
        }

        out.close();

    }
        
    return 0;
}
